//=============================================================================
// DIRECTIVE CLASS
//=============================================================================

#include "directives.h"

#include "labels.h"
#include "intmath.h"
#include "macros.h"
#include "params.h"
#include "streams.h"
#include "asmclasses.h"
#include "assemble.h"
#include "iostream"

//-----------------------------------------------------------------------------
//---= table with directives =-------------------------------------------------
//-----------------------------------------------------------------------------
const char *directive_tbl[_D_LAST] = {
    "MESSAGE", "ERROR",
    "COMMENT",
    "RADIX",
    "WARN", "NOWARN",
    "TARGET",
    "ASSIGN",
    "STACK", "BSS",
    "END",

    "R3000", "R5900",

    "IF","ELSE","ENDIF",
    "ORG","ALIGN", "ALIGN$", "ALIGN$$",
    "INCLUDE",
    "UNDEF","DEFINE","REGDEF",

    //////////////////////////////////////

    "DB","DH","DW","DD","DQ",
    "DH_UNALIGNED","DW_UNALIGNED","DD_UNALIGNED","DQ_UNALIGNED",

    "MACRO","ENDM",
    "TABLE","ENDT",
    "PROC","ENDP",
    "SEGMENT","ENDS",
    "VPUCODE", "ENDVPU",

    "EQU"
};

const char *_TEXT = "TEXT";
const char *_DATA = "DATA";

const char *_BINARY = "BINARY";
const char *_ALIGN = "ALIGN";
const char *_DUP = "DUP";

const char *_REG_PC = "pc";
const char *_REG_GP = "gp";
const char *_REG_SP = "sp";

const char *__BIN = "BIN";
const char *__OCT = "OCT";
const char *__DEC = "DEC";
const char *__HEX = "HEX";

const char _PSX_[] = "PSX";
const char _SCE_[] = "SCE";
const char _ECO_[] = "ECO";
const char _CPE_[] = "CPE";
const char _BIN_[] = "BIN";
const char _ELF_[] = "ELF";

const char _MACROMODE[] = "MACROMODE";  //argument of VPUCODE
const char _MICROMODE[] = "MICROMODE";

//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_msg(char *args, BOOL err)
{
    //if(label.if_predef()) return(RETC_SYNTAX_ERR);
    if((!err)&&(!args)) return(ERR_WARNING|RETC_PARAM_EXPECTED);
    if(assemble.preprocess) return(RETC_OK);  //only on the last processing

    PARAM param( PARAM_GET_STRINGS|PARAM_GET_VALUES );
    PARAMLIST *pl;
    DWORD ret;
    BOOL success = true;

    if(args) if((ret = param.scan(args)) != RETC_OK) return(ERR_WARNING|ret);

    if(err) error_cntr++;

    FILELIST *fl = stream.get();
    if(fl)
    {
        if(err)
        {
            if(errorFile)
            {
                errorFile << fl->name << "(" << fl->line_cntr <<
                    ") : user-generated error";
            } else {
                cout << fl->name << "(" << fl->line_cntr <<
                    ") : user-generated error";
            }
            if(args)
            {
                if(errorFile)
                {
                    errorFile << ": ";
                } else {
                    cout << ": ";
                }
            }
        } else {
            if(errorFile)
            {
                errorFile << fl->name << "(" << fl->line_cntr << ") - ";
            } else {
                cout << fl->name << "(" << fl->line_cntr << ") - ";
            }
        }
    }

    if(args) for(ret = 0; ret < param.ParamList->members; ret++)
    {
        pl = (PARAMLIST *)param.ParamList->get(ret);
        if(pl->retc != RETC_OK)
        {
            success = 0;
        } else {
            switch(pl->type)
            {
            case PTYPE_STRING:
                if(errorFile)
                {
                    errorFile << pl->str;
                } else {
                    cout << pl->str;
                }
                break;

            case PTYPE_OFF:
            case PTYPE_DEFOFF:
            case PTYPE_BASE:
                success = false;
                break;

            default:
                if (errorFile)
                {
                    errorFile << (__uint32)(pl->val);   //2do: this fuck'ng stream doesnt support __uint64
                } else {
                    cout << (__uint32)(pl->val);  // here too
                }
            }
        }
    }

    if(errorFile)
    {
        errorFile << endl;
    } else {
        cout << endl;
    }

    if(!success)
    {
        do_error(ERR_WARNING|RETC_ERR_IN_MSG, ERR_SHOW_POS);
    }

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_comment(char *args)
{
    if(label.if_predef()) return(RETC_SYNTAX_ERR);
    if(!args) return(RETC_PARAM_EXPECTED);

    char delimiter = '$';

    args = assemble.skip_nonsens(args);	//skip spaces and tabs
    if(*args) delimiter = *args;

    char *line;
    for(;;)
    {
        line = stream.in.line(STREAM_THIS_FILE);
        if(!line)
        {
            do_error(ERR_WARNING|RETC_DELIMITER_EXPECTED, ERR_SHOW_POS);
            break;
        }
        if(strchr(line, delimiter)) break;
    }

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_set_radix(char *val)
{
    if(!val) return(RETC_PARAM_EXPECTED);
    dbgout("DIRECTIVE::do_set_radix(CHAR[" << val << "])\n")

        val = assemble.skip_nonsens(val);
    if(*assemble.split_str(val)) return(RETC_SYNTAX_ERR);

    if(assemble.cmp(val, __BIN)) intmath.def_base = IMO_BASE_BIN;
    else if(assemble.cmp(val, __OCT)) intmath.def_base = IMO_BASE_OCT;
    else if(assemble.cmp(val, __DEC)) intmath.def_base = IMO_BASE_DEC;
    else if(assemble.cmp(val, __HEX)) intmath.def_base = IMO_BASE_HEX;
    else return(RETC_SYNTAX_ERR);

    //label.preset(assemble.prog_cntr, LBL_TYPE_UNDEF);

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_set_warnlevel(char *args, BOOL state)
{
    dbgout("DIRECTIVE::do_set_warnlevel(..)\n")
    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_sel_target(char *args)
{
    dbgout("DIRECTIVE::do_sel_target(..)\n")

    if(!args)
    {
        return(RETC_PARAM_EXPECTED);
    }

    BOOL iset = false;
    BOOL tset = false;

    for(;;)
    {
        args = assemble.skip_nonsens(args);	//skip spaces and tabs
        if(*args == NULL) return(RETC_PARAM_EXPECTED);

        if(assemble.cmp(args, _PSX_, 3))
        {
            if(tset) return(RETC_TOO_MUCH_PARAMS);
            if(assemble.targ_sel) do_error(RETC_TARG_ALRD_SEL|ERR_WARNING, ERR_SHOW_POS);
            else {assemble.targ_sel = true; assemble.binary = false; assemble.config.target = SEL_TARG_PSX;}
            tset = true;
            args+=3;
        } else if(assemble.cmp(args, _SCE_, 3)) {
            if(tset) return(RETC_TOO_MUCH_PARAMS);
            if(assemble.targ_sel) do_error(RETC_TARG_ALRD_SEL|ERR_WARNING, ERR_SHOW_POS);
            else {assemble.targ_sel = true; assemble.binary = false; assemble.config.target = SEL_TARG_SCE;}
            tset = true;
            args+=3;
        } else if(assemble.cmp(args, _ECO_, 3)) {
            if(tset)
            {
                return(RETC_TOO_MUCH_PARAMS);
            }
            if(assemble.targ_sel)
            {
                do_error(RETC_TARG_ALRD_SEL|ERR_WARNING, ERR_SHOW_POS);
            } else {
                assemble.targ_sel = true;
                assemble.binary = false;
                assemble.config.target = SEL_TARG_ECO;
            }
            tset = true;
            args+=3;
        } else if(assemble.cmp(args, _CPE_, 3)) {
            if(tset)
            {
                return(RETC_TOO_MUCH_PARAMS);
            }
            if(assemble.targ_sel)
            {
                do_error(RETC_TARG_ALRD_SEL|ERR_WARNING, ERR_SHOW_POS);
            } else {
                assemble.targ_sel = true;
                assemble.binary = false;
                assemble.config.target = SEL_TARG_CPE;
            }
            tset = true;
            args+=3;
        } else if(assemble.cmp(args, _BIN_, 3)) {
            if(tset)
            {
                return(RETC_TOO_MUCH_PARAMS);
            }
            if(assemble.targ_sel)
            {
                do_error(RETC_TARG_ALRD_SEL|ERR_WARNING, ERR_SHOW_POS);
            }   else {
                assemble.targ_sel = true;
                assemble.binary = true;
                assemble.config.target = SEL_TARG_BIN;
            }
            tset = true;
            args+=3;
        } else if(assemble.cmp(args, _ELF_, 3)) {
            if(tset) return(RETC_TOO_MUCH_PARAMS);
            if(assemble.targ_sel) do_error(RETC_TARG_ALRD_SEL|ERR_WARNING, ERR_SHOW_POS);
            else {assemble.targ_sel = true; assemble.binary = false; assemble.config.target = SEL_TARG_ELF;}
            tset = true;
            args+=3;
        } else if(assemble.cmp(args, directive_tbl[D_R3000], 5)) {
            if(iset) return(RETC_TOO_MUCH_PARAMS);
            if(
                (assemble.config.iset == SEL_ISET_VPU0)||
                (assemble.config.iset == SEL_ISET_VPU1)
                )
            {
                do_error(RETC_ISET_LOCKED|ERR_WARNING, ERR_SHOW_POS);
            } else {
                assemble.config.iset = SEL_ISET_R3000;
            }
            iset = true;
            args+=5;
        } else if(assemble.cmp(args, directive_tbl[D_R5900], 5)) {
            if(iset) return(RETC_TOO_MUCH_PARAMS);
            if(
                (assemble.config.iset == SEL_ISET_VPU0)||
                (assemble.config.iset == SEL_ISET_VPU1)
                )
            {
                do_error(RETC_ISET_LOCKED|ERR_WARNING, ERR_SHOW_POS);
            } else {
                assemble.config.iset = SEL_ISET_R5900;
            }
            iset = true;
            args+=5;
        } else {
            return(RETC_SYNTAX_ERR);
        }

        args = assemble.skip_nonsens(args);	//skip spaces and tabs
        switch(*args)
        {
            case NULL: return(RETC_OK);
            case ',': args++; break;
            default: return(RETC_PARAM_EXPECTED);
        }
    }
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_assign(char *args)
{
    if(!args) return(RETC_PARAM_EXPECTED);
    dbgout("DIRECTIVE::do_assign(..)\n")

        IM_RET imr;
    PARAM param( PARAM_GET_STRINGS );
    DWORD pnum = 0;
    PARAMLIST *pl;
    DWORD ret;


    if((ret = param.scan(args)) != RETC_OK) return(ret);
    while(pnum < param.ParamList->members)
    {
        pl = (PARAMLIST*)param.ParamList->get(pnum++);
        if(!pl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);

        args = pl->str;

        ret=0;
        if(assemble.cmp(args, _REG_PC, 2)) {args+=2; ret=1;}
        else if(assemble.cmp(args, _REG_GP, 2)) {args+=2; ret=2;}
        else if(assemble.cmp(args, _REG_SP, 2)) {args+=2; ret=3;}

        while((*args=='\t')||(*args==' ')) args++;	//skip spaces and tabs
        if(*args == ':') args++; else ret=0;

        imr.str = args;
        if(!intmath.calc(&imr)) do_error(imr.imrc, ERR_SHOW_POS);
        else switch(ret)
        {
            case 1:
                if((assemble.targ_is_entry)&&(assemble.targ_entry == imr.value))
                    do_error(ERR_WARNING|RETC_OVRWR_ASSIGN, ERR_SHOW_POS);
                assemble.targ_entry=imr.value; assemble.targ_is_entry=true;
                break;
            case 2:
                if((assemble.targ_is_gp)&&(assemble.targ_gp == imr.value))
                    do_error(ERR_WARNING|RETC_OVRWR_ASSIGN, ERR_SHOW_POS);
                assemble.targ_gp=imr.value; assemble.targ_is_gp=true;
                break;
            case 3:
                if((assemble.targ_is_stack)&&(assemble.targ_stack == imr.value))
                    do_error(ERR_WARNING|RETC_OVRWR_ASSIGN, ERR_SHOW_POS);
                assemble.targ_stack=imr.value; assemble.targ_is_stack=true;
                break;

            default: do_error(RETC_SYNTAX_ERR, ERR_SHOW_POS);
        }
    }

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_stack_bss(char *args, char type)
{

    if(label.if_predef()) return(RETC_SYNTAX_ERR);
    if(!args) return(RETC_PARAM_EXPECTED);

    IM_RET imr;
    imr.str = args;
    if(!intmath.calc(&imr)) do_error(imr.imrc, ERR_SHOW_POS);

    if(type)  //BSS
    {
        if(!label.add("BSS", imr.value, LBL_TYPE_DB)) return(RETC_OK);

        if(!assemble.binary)
        {
            if((assemble.targ_is_bss) && (assemble.targ_bss != imr.value))
                do_error(ERR_WARNING|RETC_OVRWR_ASSIGN, ERR_SHOW_POS);
            assemble.targ_bss = imr.value;
            assemble.targ_is_bss = true;
        }

    }
    else	//STACK
    {
        imr.value = imr.value-imr.value%4;
        if(!label.add("STACK", imr.value, LBL_TYPE_DW)) return(RETC_OK);

        if(!assemble.binary)
        {
            if((assemble.targ_is_stack) && (assemble.targ_stack != imr.value))
                do_error(ERR_WARNING|RETC_OVRWR_ASSIGN, ERR_SHOW_POS);
            assemble.targ_stack = imr.value;
            assemble.targ_is_stack = true;
        }

    }

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_eof(char *args)
{
    dbgout("DIRECTIVE::do_eof(..)\n")

        IM_RET imr;

    assemble.end_of_file=1;

    if(assemble.binary)
    {
        if(args) return(RETC_TOO_MUCH_PARAMS);
    }
    else if(args)
    {
        imr.str = args;
        if(!intmath.calc(&imr)) do_error(imr.imrc, ERR_SHOW_POS);
        if((assemble.targ_is_entry) && (assemble.targ_entry != imr.value))
            return(RETC_OVRWR_ASSIGN);
        assemble.targ_entry = imr.value;
        assemble.targ_is_entry = true;
    }

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_if(char *args)
{
    if(label.if_predef()) return(RETC_SYNTAX_ERR);
    if(!args) return(RETC_PARAM_EXPECTED);

    dbgout("DIRECTIVE::do_if(CHAR[" << args << "])\n")

        IM_RET imr;
    imr.str = args;
    //if(!intmath.calc(&imr, ITYPE_NOLABELS)) return(imr.imrc);
    if(!intmath.calc(&imr, ITYPE_LOGICAL)) do_error(imr.imrc, ERR_SHOW_POS);
    //return(imr.imrc);

    FILELIST *fl, *fl_src;
    fl_src = stream.get();
    DWORD line_cntr = (fl_src)?(fl_src->line_cntr):(0);

    char *line, *tail;
    BOOL record = (imr.value)?1:0;	//record to memory

    char *code;
    if(imr.imrc == RETC_OK)
    {
        if(!(code = (char *)malloc(1))) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
        code[0] = NULL;
    }
    else
    {
        code = NULL;
    }

    DWORD new_size = 0, size = 0, num_lines = 0;
    WORD if_cntr = 0;

    //read lines into memory
    //IF..(record)..ELSE/ENDIF
    for(;;)
    {
        line = stream.in.line(STREAM_THIS_FILE);
        if(!line)
        {
            do_error(RETC_ENDIF_EXPECTED, ERR_SHOW_POS);
            if(code) free(code);
            return(RETC_OK);
        }
        assemble.detect_end(line);
        line = assemble.skip_nonsens(line);

        new_size = size;
        if((code)&&(record))
        {	//add 2 code
            new_size += strlen(line) + 1;
            code = (char *)realloc(code, new_size);
            if(!code) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
            strcpy(&code[size], line);
        }

        tail = assemble.split_str(line);
        if(assemble.cmp(line, directive_tbl[D_IF]))
        {
            if_cntr++;
            goto nxt_line;
        }
        if(assemble.cmp(line, directive_tbl[D_ENDIF]))
        {
            if(if_cntr)
            {
                if_cntr--;
                goto nxt_line;
            }

            if((tail)&&(*tail)) do_error(ERR_WARNING|RETC_EXTRA_CHAR, ERR_SHOW_POS);
            goto exec_code;
        }
        if(assemble.cmp(line, directive_tbl[D_ELSE]))
        {
            if(if_cntr)	goto nxt_line;

            if((tail)&&(*tail)) do_error(ERR_WARNING|RETC_EXTRA_CHAR, ERR_SHOW_POS);
            new_size = size;
            record = record ^1;
            if((code)&&(record))
            {	//update line for else block
                fl_src = stream.get();
                line_cntr = (fl_src)?(fl_src->line_cntr):(0);
            }
        }

nxt_line:
        size = new_size;
        if(record) num_lines++;
    }


exec_code:
    if(!code) return(RETC_OK);

    code = (char *)realloc(code, size+1);
    if(!code) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
    code[size] = NULL;

    //add to stream
    if(stream.get())
    {
        fl = stream.add("", FTYPE_IFELSE|FTYPE_BUF);
        if(!(fl_src = stream.get(fl->last))) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);

        fl->name = (char *)realloc(fl->name, strlen(fl_src->name)+32);
        if(!fl->name) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
        if(imr.value)
            sprintf(fl->name, "%s(%i):IF", fl_src->name, line_cntr);
        else
            sprintf(fl->name, "%s(%i):ELSE", fl_src->name, line_cntr);
    }
    else
    {
        if(imr.value)
            fl = stream.add("IF", FTYPE_IFELSE|FTYPE_BUF);
        else
            fl = stream.add("ELSE", FTYPE_IFELSE|FTYPE_BUF);
    }

    fl->buf = fl->pos = code;
    fl->size = num_lines;

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_org(char *val)
{
    dbgout("DIRECTIVE::do_org(CHAR[" << val << "])\n")
        IM_RET imr;

    imr.str=val;
    if(!intmath.calc(&imr, ITYPE_NOLABELS)) return(imr.imrc);
    //no undefined labels allowed (in do_math)
    //cause we need to change all following addresses

    if(imr.value<0) return(RETC_CODE_OVRFLW);
    assemble.prog_cntr = imr.value;
    //label.preset(assemble.prog_cntr, LBL_TYPE_UNDEF);

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_align(char *val, BOOL fofs)
{
    dbgout("DIRECTIVE::do_align(CHAR[" << val << "])\n")
        IM_RET imr;

    imr.str=val;
    if(!intmath.calc(&imr, ITYPE_NOLABELS)) return(imr.imrc);
    //no undefined labels allowed (in do_math)
    //cause we need to change all following addresses

    if((imr.value<1)||(imr.value>0xffff)) return(RETC_CODE_OVRFLW);

    if(fofs) //then align to fileoffset
    {
        while(stream.out.fofs % imr.value) stream.out.opcode(NULL, 1);
    }
    else
    {
        stream.out.align(imr.value);
    }

    //label.preset(assemble.prog_cntr, LBL_TYPE_UNDEF);

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_include(char *args)
{
    if(!args) return(RETC_PARAM_EXPECTED);
    dbgout("DIRECTIVE::do_include(CHAR[" << args << "])\n")

        BOOL binary=false;

    while((*args=='\t')||(*args==' ')) args++;					//skip spaces and tabs
    if(assemble.cmp(args, _BINARY, 6)) {args+=6; binary=true;}
    while((*args=='\t')||(*args==' ')) args++;					//skip spaces and tabs

    if(*args!='"') return(RETC_SYNTAX_ERR);						//filename begins with a mark
    char *scan=++args;											//test filename-characters
    while((*scan>=32)&&(*scan!='*')&&(*scan!='?')&&(*scan!='<')
        &&(*scan!='>')&&(*scan!='|')&&(*scan!='"')) scan++;
    if(*scan!='"') return(RETC_SYNTAX_ERR);						//filename ends with a mark
    *(scan++)=NULL;
    while((*scan=='\t')||(*scan==' ')) scan++;					//skip spaces and tabs
    if(*scan)													//check for align
    {
        if(!assemble.cmp(scan, _ALIGN, 5)) return(RETC_SYNTAX_ERR);
        do_align(scan+5);
    }

    //add file to input-list
    stream.add(args, FTYPE_INCLUDE);

    DWORD ret;

    if(binary)
    {
        label.preset(assemble.prog_cntr, LBL_TYPE_DB);

        if((ret = stream.out.file()) != RETC_OK) do_error(ret, ERR_SHOW_POS);
    }

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_undef(char *args)
{
    if(!args) return(RETC_PARAM_EXPECTED);
    dbgout("DIRECTIVE::do_undef(CHAR[" << args << "])\n")

        //before definig a new label, lets give this one peace
        label.preset(assemble.prog_cntr, LBL_TYPE_UNDEF);

    while((*args==' ')||(*args=='\t')) args++;
    if(!*args) return(RETC_PARAM_EXPECTED);
    int bkc = strlen(args)-1;
    while((bkc>0)&&((args[bkc]==' ')||(args[bkc]=='\t'))) bkc--;
    args[bkc+1] = NULL;

    label.add(args, 0, LBL_TYPE_DEFINE);
    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_define(char *args, BYTE type)
{
    if(!args) return(RETC_PARAM_EXPECTED);
    dbgout("DIRECTIVE::do_define(CHAR[" << args << "], DWORD[" << type << "])\n")

        IM_RET imr;

    //before definig a new label, lets give this one peace
    label.preset(assemble.prog_cntr, LBL_TYPE_UNDEF);

    while((*args==' ')||(*args=='\t')) args++;
    char *val=args;
    while((*val)&&(*val!=' ')&&(*val!='\t')&&(*val!='=')) val++;
    if(*val) *(val++) = NULL;
    while((*val==' ')||(*val=='\t')) val++;

    if(!*val)
    {
        //return(RETC_PARAM_EXPECTED);
        imr.value = -1;	//define as true by default
    }
    else if(*val != '=') return(RETC_SYNTAX_ERR);
    else
    {
        imr.str = ++val;
        if(!intmath.calc(&imr)) return(imr.imrc);
    }

    label.add(args, imr.value, type);
    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_equ(char *lbl, char *args)
{
    if((!lbl)||(!args)) return(RETC_PARAM_EXPECTED);
    dbgout("DIRECTIVE::do_equ(CHAR[" << lbl << "], CHAR[" << args << "])\n")

        //before definig a new label, lets give this one peace
        label.preset(assemble.prog_cntr, LBL_TYPE_UNDEF);

    IM_RET imr;
    imr.str = args;
    if(!intmath.calc(&imr)) return(imr.imrc);

    label.add(lbl, imr.value, LBL_TYPE_EQUATE);
    return(RETC_OK);
}//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_dup(char *args, int base)
{
    if(!args) return(RETC_PARAM_EXPECTED);
    //support: Dx <value> DUP(xxx)
    IM_RET imr;

    WORD last = strlen(args);
    if((last<2)||(args[--last] != ')')) return(RETC_SYNTAX_ERR);
    args[last] = NULL;

    int brkt = 1;
    int marks = 0;
    for(last--;(last)&&(brkt);last--) switch(args[last])
    {
        case '\'': if(!marks) marks=1; else if(marks==1) marks=0; break;
        case '"': if(!marks) marks=2; else if(marks==2) marks=0; break;
        case ')': if(!marks) brkt++; break;
        case '(': if(!marks) brkt--; break;
    }
    if((brkt != 0)||(args[last+1] != '(')) return(RETC_SYNTAX_ERR);
    char *val = args+last+1;
    *(val++) = NULL;
    last-=2;
    if(!assemble.cmp(args+last, _DUP)) return(RETC_SYNTAX_ERR);

    args[last] = NULL;
    imr.str = args;
    if(!intmath.calc(&imr, ITYPE_NOLABELS)) return(imr.imrc);
    if((imr.value < 1)||(imr.value > 0xffff)) return(RETC_CODE_OVRFLW);
    WORD dup = (WORD)imr.value;

    if(assemble.preprocess)
    {	//just create placeholder
        imr.value = 0;
    }
    else
    {
        imr.str = val;
        if(!intmath.calc(&imr)) return(imr.imrc);
    }

    for(;dup;dup--) stream.out.opcode(imr.value, base);

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_data(char *lbl, char *args, int base, int align)
{
    if(!args) return(RETC_PARAM_EXPECTED);
    dbgout("DIRECTIVE::do_data(CHAR[" << args << "], INT[" << base << "])\n")

        //do label stuff
        BYTE lbl_type;

    if (align <= 0)
        align = base;
    stream.out.align(align);
    switch(base)
    {
        case 1: lbl_type=LBL_TYPE_DB; break;
        case 2:	lbl_type=LBL_TYPE_DH; break;
        case 4: lbl_type=LBL_TYPE_DW; break;
        case 8: lbl_type=LBL_TYPE_DD; break;
        case 16: lbl_type=LBL_TYPE_DQ; break;
        default: lbl_type=LBL_TYPE_UNDEF;
    }
    if(label.if_predef())
    {
        if(lbl) return(RETC_SYNTAX_ERR);
        label.preset(assemble.prog_cntr, lbl_type);
    }
    else if(lbl)
    {
        label.add(lbl, assemble.prog_cntr, lbl_type);
    }

    PARAM param( PARAM_GET_VALUES|PARAM_GET_STRINGS );
    DWORD pnum = 0;
    PARAMLIST *pl;
    DWORD ret;
    char *str;

    if((ret = param.scan(args)) != RETC_OK) return(ret);
    while(pnum < param.ParamList->members)
    {
        pl = (PARAMLIST*)param.ParamList->get(pnum++);
        if(!pl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);
        switch(pl->retc)
        {
            case RETC_OK:
                switch(pl->type)
                {
                    case PTYPE_UNDEF:
                    case PTYPE_VALUE:
                    case PTYPE_REG:
                    case PTYPE_LBL:
                        stream.out.opcode(pl->val, base);
                        //stream.out.align(align);
                        break;

                    case PTYPE_STRING:
                        for(str = pl->str; *str; str++)
                        {
                            stream.out.opcode(*str, 1);
                            stream.out.align(align);
                        }
                        break;

                    default:
                        do_error(RETC_SYNTAX_ERR, ERR_SHOW_POS); //error in parameters???
                }
                break;

            case RETC_SYNTAX_ERR:
                if((ret = do_dup(pl->str, base)) != RETC_OK)
                    do_error(ret, ERR_SHOW_POS);
                break;

            default:
                do_error(pl->retc, ERR_SHOW_POS); //error in parameters???
        }
    }

    return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_seg(char *lbl, char *args)
{
    IM_RET imr;
    if(!lbl) return(RETC_PARAM_EXPECTED);
    if(!label._validate(lbl)) return(RETC_UNKWN_LBL_SYNTAX);

    imr.str = args;
    if(!intmath.calc(&imr, ITYPE_NOLABELS)) return(imr.imrc);

    assemble.prog_cntr = imr.value;
    //label.preset(assemble.prog_cntr, LBL_TYPE_UNDEF);

    return( asmclass.add(lbl, CLS_TYPE_SEG) );
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_class(char *lbl, char *args, char type)
{
    IM_RET imr;
    if(!lbl) return(RETC_PARAM_EXPECTED);
    if(!label._validate(lbl)) return(RETC_UNKWN_LBL_SYNTAX);

    //do alignment (check args)
    if(args)
    {
        while((*args=='\t')||(*args==' ')) args++;
        if(assemble.cmp(args, _ALIGN, 5))
        {
            args+=5;
            if((*args!='\t')&&(*args!=' ')) return(RETC_SYNTAX_ERR);
            args++;
            imr.str = args;
            if(!intmath.calc(&imr, ITYPE_NOLABELS)) return(imr.imrc);
            if((imr.value<0)||(imr.value>0xffff)) return(RETC_OVERFLOW);
            stream.out.align((WORD)imr.value);
        }
        else
        {
            //if(*args) return(RETC_SYNTAX_ERR);
            stream.out.align(4);	//4 by default
        }
    }
    else stream.out.align(4);	//4 by default

    //label.preset(assemble.prog_cntr, LBL_TYPE_CODE);

    return( asmclass.add(lbl, type) );
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_end(char *lbl, char *args, char type)
{
    if(!lbl)
    {
        return(RETC_PARAM_EXPECTED);
    }
    if(args)
    {
        return(RETC_TOO_MUCH_PARAMS);
    }

    return( asmclass.close(lbl, type) );
}
//-----------------------------------------------------------------------------
DWORD
DIRECTIVE::do_vpucode(char *lbl, char *args)
{
    // 0 = macro_mode, 1 = micro_mode
    char mode;

    if(
        (assemble.config.iset == SEL_ISET_VPU0) ||
        (assemble.config.iset == SEL_ISET_VPU1)
        )
    {
        return(RETC_ENDVPU_EXPECTED);
    }
    if( assemble.config.iset != SEL_ISET_R5900 )
    {
        return(RETC_TARG_MISMATCH);
    }

    if( !lbl )
    {
        return(RETC_PARAM_EXPECTED);
    }

    if(!args)
    {
        mode = 1;
    } else {
        args = assemble.skip_nonsens(args);
        if(assemble.cmp(args, "0"))
        {
            mode = 0;
        } else if(assemble.cmp(args, "1")) {
            mode = 1;
        } else if(assemble.cmp(args, _MACROMODE)) {
            mode = 0;
        } else if(assemble.cmp(args, _MICROMODE)) {
            mode = 1;
        } else {
            return(RETC_SYNTAX_ERR);
        }
    }

    if(mode==0) stream.out.align(4);  //align for 32bit instructions
    else stream.out.align(8);         //align for 64bit instructions

    //add to class-table
    DWORD ret = asmclass.add(lbl, CLS_TYPE_VPU);

    //set instr set to vpu
    if(ret == RETC_OK)
    {
        if( mode==0 )
        {
            assemble.config.iset = SEL_ISET_VPU0;
        } else {
            assemble.config.iset = SEL_ISET_VPU1;
        }
    }

    return( ret );
}
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::do_endvpu(char *lbl, char *args)
{
    if(!lbl) return(RETC_PARAM_EXPECTED);
    if(args) return(RETC_TOO_MUCH_PARAMS);

    if((assemble.config.iset != SEL_ISET_VPU0) &&
        (assemble.config.iset != SEL_ISET_VPU1)) return(RETC_VPUCODE_EXPECTED);

    //delete from class table
    DWORD ret = asmclass.close(lbl, CLS_TYPE_VPU);

    //return to R5900 instr set
    /*if(ret == RETC_OK)*/ assemble.config.iset = SEL_ISET_R5900;

    return( ret );
}
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// DIRECTIVE::process_directive: do the things the directives should do
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::process_directive(char *succ, int dirc, char *args)
{
    dbgout(
        "DIRECTIVE::process_directive(CHAR*[" << (DWORD)succ << "], INT[" <<
        dirc << "], CHAR[" << args << "])\n"
        )

    if( (succ) && (!(*succ)) )
    {
        cout << "succ nulling: " << succ << endl;
        succ = NULL;
    }
    if( (args)&&(!(*args)) )
    {
        args = NULL;
    }

    //don't allow 2 labels per line
    if(label.if_predef())
    {
        if(succ) return(RETC_SYNTAX_ERR);
    }
    //	else if((succ) && (dirc < D_MACRO))
    //	{
    //		if(!label.predef(succ)) return(RETC_OK);
    //	}

    if((succ) && (dirc < D_DB))
    {
        return(RETC_SYNTAX_ERR);
    }

    //write listing
    if((!assemble.preprocess)&&(assemble.lst_fh))
    {
        switch(dirc)
        {
            case D_ORG:
            case D_ALIGN:
            case D_ALIGN$:
            case D_ALIGN$$:
            case D_DB:
            case D_DH:
            case D_DW:
            case D_DD:
            case D_DH_UNALIGNED:
            case D_DW_UNALIGNED:
            case D_DD_UNALIGNED:
#ifdef __GNUC__
                fprintf(assemble.lst_fh, "%08llX\t",	assemble.prog_cntr);
#else
                fprintf(assemble.lst_fh, "%08I64X\t",	assemble.prog_cntr);
#endif
                if(succ)
                {
                    fprintf(assemble.lst_fh, "%s ", succ);
                }
                fprintf(assemble.lst_fh, "%s ",	directive_tbl[dirc]);
                if(args)
                {
                    fprintf(assemble.lst_fh, "%s",	args);
                }
                fprintf(assemble.lst_fh, "\n");
        }
    }

    switch(dirc)
    {
        case D_MESSAGE:
            return(do_msg(args, false));
        case D_ERROR:
            return(do_msg(args, true));
        case D_COMMENT:
            return(do_comment(args));
        case D_RADIX:
            return(do_set_radix(args));
        case D_WARN:
            return(do_set_warnlevel(args, true));
        case D_NOWARN:
            return(do_set_warnlevel(args, false));
        case D_TARGET:
            return(do_sel_target(args));
        case D_ASSIGN:
            if(assemble.binary)
            {
                return(RETC_TARG_MISMATCH);
            } else {
                return(do_assign(args));
            }
        case D_STACK:
            return(do_stack_bss(args, 0));
        case D_BSS:
            return(do_stack_bss(args, 1));
        case D_END:
            return(do_eof(args));
        case D_R3000:
            if(args) return(RETC_TOO_MUCH_PARAMS);
            if(
                (assemble.config.iset == SEL_ISET_VPU0)||
                (assemble.config.iset == SEL_ISET_VPU1)
                )
            {
                do_error(RETC_ISET_LOCKED|ERR_WARNING, ERR_SHOW_POS);
            } else {
                assemble.config.iset = SEL_ISET_R3000;
            }
            break;
        case D_R5900:
            if(args)
            {
                return(RETC_TOO_MUCH_PARAMS);
            }
            if(
                (assemble.config.iset == SEL_ISET_VPU0)||
                (assemble.config.iset == SEL_ISET_VPU1)
                )
            {
                do_error(RETC_ISET_LOCKED|ERR_WARNING, ERR_SHOW_POS);
            } else {
                assemble.config.iset = SEL_ISET_R5900;
            }
            break;
        case D_IF:
            return(do_if(args));
        case D_ELSE:
            return(RETC_ELSE_WO_IF);
        case D_ENDIF:
            return(RETC_ENDIF_WO_IF);
        case D_ORG:
            return(do_org(args));			//set pc_cntr to value(args)
        case D_ALIGN:
        case D_ALIGN$:
            return(do_align(args));		//send output till pc_cntr % value(args) = 0
        case D_ALIGN$$:
            if(!assemble.binary)
            {
                return(RETC_TARG_MISMATCH);
            } else {
                return(do_align(args,true));
            }
        case D_INCLUDE:
            return(do_include(args));
            //args=BINARY then copy file to output else update file-list
        case D_UNDEF:
            return(do_undef(args));
        case D_DEFINE:
            return(do_define(args, LBL_TYPE_DEFINE));	//set label with value args
        case D_REGDEF:
            return(do_define(args, LBL_TYPE_REG));		//set label with value args
        case D_DB:
            return(do_data(succ, args, 1));	//include data (1,2,4,8,16 bytes)
        case D_DH:
            return(do_data(succ, args, 2));	//put data into packets and write
        case D_DH_UNALIGNED:
            return(do_data(succ, args, 2, 1));
        case D_DW:
            return(do_data(succ, args, 4));
        case D_DW_UNALIGNED:
            return(do_data(succ, args, 4, 1));
        case D_DD:
            return(do_data(succ, args, 8));
        case D_DD_UNALIGNED:
            return(do_data(succ, args, 8, 1));
            //example DH "abc123", "ABC123", 5*(3-7) ;extends every character to a HW
        //case D_DQ: return(do_data(succ, args, 16));
        //case D_DQ_UNALIGNED: return(do_data(succ, args, 16, 1));
        case D_MACRO:
            return(macro.add(succ, args));
        case D_ENDM:
            if(!succ)
            {
                return(RETC_PARAM_EXPECTED);
            } else {
                return(RETC_MACRO_EXPECTED);
            }
        case D_TABLE:
            return(do_class(succ, args, CLS_TYPE_TBL));
        case D_ENDT:
            return(do_end(succ, args, CLS_TYPE_TBL));
        case D_PROC:
            return(do_class(succ, args, CLS_TYPE_PROC));
        case D_ENDP:
            return(do_end(succ, args, CLS_TYPE_PROC));
        case D_SEGMENT:
            return(do_seg(succ, args));
        case D_ENDS:
            return(do_end(succ, args, CLS_TYPE_SEG));
        case D_VPUCODE:
            return(do_vpucode(succ, args));
        case D_ENDVPU: return(do_endvpu(succ, args));
        case D_EQU: return(do_equ(succ, args));
        default:
            return(RETC_UNKWN_INSTR);
    }
    return(RETC_OK);
}

//-----------------------------------------------------------------------------
// DIRECTIVE::exec: scan the string for a valid directive
//-----------------------------------------------------------------------------
DWORD DIRECTIVE::exec(char *succ, char *dirc, char *args)
{
    dbgout(
        "DIRECTIVE::exec(CHAR*[" << (DWORD)succ << "], CHAR[" << dirc <<
        "], CHAR[" << args << "])\n"
        )
    int scan_cnt = 0;
    //if(succ) scan_cnt = DIRECTIVES_SUCC;
    //no succ? then search only 4 predecessor directives (jump over succ)

    for(; scan_cnt < _D_LAST; scan_cnt++)
    {
        if(assemble.cmp(dirc, directive_tbl[scan_cnt]))
        {
            return(process_directive(succ, scan_cnt, args));
        }
    }
    return(RETC_UNKWN_INSTR);
}

//-----------------------------------------------------------------------------
// DIRECTIVE::if_exist: test if string is directive
//-----------------------------------------------------------------------------
BOOL DIRECTIVE::if_exist(char *dirc)
{
    for(int scan_cnt=0; scan_cnt<_D_LAST; scan_cnt++)
    {
        if(assemble.cmp(dirc, directive_tbl[scan_cnt]))
        {
            return(true);
        }
    }
    return(false);
}
